const describeMimeTypeArray = mimeTypeArray => (
  Array.from(mimeTypeArray).map(mimeType => ({
    description: mimeType.description,
    enabledPlugin: mimeType.enabledPlugin.filename,
    suffixes: mimeType.suffixes,
    type: mimeType.type,
  }))
);


const describePluginArray = pluginArray => (
  Array.from(pluginArray).map(plugin => ({
    description: plugin.description,
    filename: plugin.filename,
    length: plugin.length,
    name: plugin.name,
    mimeTypes: Array.from(plugin).map(({ type }) => type),
    version: plugin.version,
  }))
);


function describeValue(value, argumentArrays = [[]]) {
  const type = typeof value;
  const description = { type };

  if (['boolean', 'number', 'string', 'undefined'].includes(type)) {
    description.value = value;
  } else if (value === null) {
    description.type = 'null';
    description.value = null;
  } else if (type === 'function') {
    argumentArrays.forEach((argumentArray) => {
      let error;
      let returnValue;
      try {
        returnValue = value.apply(this, argumentArray);
      } catch (e) {
        error = {
          message: e.message,
          name: e.name,
        };
      }
      description.value = [{
        arguments: argumentArray,
        error,
        value: describeValue(returnValue),
      }];
    });
  } else if (window.MimeTypeArray && value instanceof window.MimeTypeArray) {
    description.type = 'MimeTypeArray';
    description.value = describeMimeTypeArray(value);
  } else if (window.PluginArray && value instanceof window.PluginArray) {
    description.type = 'PluginArray';
    description.value = describePluginArray(value);
  } else {
    description.value = value.toString();
  }

  return description;
}


const getNavigatorFingerprint = () => {
  const descriptors = Object.getOwnPropertyDescriptors(window.navigator);
  const prototype = Object.getPrototypeOf(window.navigator);
  const prototypeDescriptors = Object.getOwnPropertyDescriptors(prototype);

  [descriptors, prototypeDescriptors].forEach((currentDescriptors) => {
    Object.keys(currentDescriptors)
      .forEach((key) => {
        const value = describeValue.call(window.navigator, window.navigator[key]);

        if (currentDescriptors[key].get) {
          // eslint-disable-next-line no-param-reassign
          currentDescriptors[key].get = value;
        } else {
          // eslint-disable-next-line no-param-reassign
          currentDescriptors[key].value = value;
        }
      });
  });
  return {
    descriptors,
    prototypeDescriptors,
  };
};


export default getNavigatorFingerprint;
